#include <inttypes.h>
#define xmalloc malloc
struct en_tree_st
{
  char *name;
  uintmax_t size;
  struct en_tree_st *parent, *child, *next, *prev;
  int r, g, b;
};

typedef struct en_tree_st en_tree;

/* Initialize a node to safe defaults */

int en_tree_init(en_tree *target)
{
  if (target==NULL) return -1;
  target->size=0;
  target->next=NULL;
  target->prev=NULL;
  target->parent=NULL;
  target->child=NULL;
  target->name=NULL;
  target->r=random()%65535;
  target->g=random()%65535;
  target->b=random()%65535;
  return 0;
}

/* Create and return a new, initialized node */

en_tree *en_tree_new()
{
  en_tree *new_tree;
  new_tree=(en_tree *) xmalloc(sizeof(en_tree));
  en_tree_init(new_tree);
  return new_tree;
}

/* Find the end of the list */

en_tree *en_tree_seek_end(en_tree *cur)
{
en_tree *current;
  for (current=cur; current->next!=NULL; current=current->next);
  return current;
}

/* Add a child to a node */
int en_tree_add_child(en_tree *target, en_tree *child)
{ 
  if (target==NULL) return -1;
  if (target->child==NULL) target->child=child;
  else
  {
    en_tree *end;
    end=en_tree_seek_end(target->child);
    end->next=child;
    child->prev=end;
  }
  child->parent=target;
  return 0;
}

/* Recursively destroy a tree branch */

void en_tree_rdestroy(en_tree *target)
{
  en_tree *cur;
  for (cur=target->child; cur!=NULL; cur=cur->next)
  {
    en_tree_rdestroy(cur);
  }
  free(target->name);
  free(target);
}

/* Remove a node from the tree without deleting it */

int en_tree_remove(en_tree *target)
{
  if (target==NULL) return -1;

// remove reference in the previous node
  if (target->prev!=NULL) target->prev->next=target->next;

// remove reference in the next node
  if (target->next!=NULL) target->next->prev=target->prev;

// remove reference in the parent node
  if (target->parent!=NULL && target->parent->child==target) 
    target->parent->child=target->next;
}

/* Set the name and size of an en_tree node */

int en_tree_set(en_tree *target, uintmax_t n_blocks, const char *name)
{
  target->size=n_blocks;
  if (target->name!=NULL) free (target->name);
  if (name==NULL) target->name=NULL;
  else target->name=strdup(name);
}

/* Set the name and size of an en_tree node */
int en_tree_add(en_tree *target, uintmax_t n_blocks, const char *name)
{
  target->size+=n_blocks;
  if (target->name!=NULL) free (target->name);
  if (name==NULL) target->name=NULL;
  else target->name=strdup(name);
}

void en_tree_incr_up(en_tree *target, uintmax_t n_blocks)
{
  if (target==NULL) return;
  for (en_tree *cur=target->parent; cur!=NULL; cur=cur->parent)
  {
	  cur->size+=n_blocks;
  }
}
